@@ -3,8 +3,8 @@
 /*
  * mcd-account.h - the Telepathy Account D-Bus interface (service side)
  *
- * Copyright (C) 2008 Collabora Ltd. <http://www.collabora.co.uk/>
- * Copyright (C) 2008 Nokia Corporation
+ * Copyright © 2008–2011 Collabora Ltd. <http://www.collabora.co.uk/>
+ * Copyright © 2008 Nokia Corporation
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -39,8 +39,6 @@ typedef struct _McdAccount McdAccount;
 typedef struct _McdAccountPrivate McdAccountPrivate;
 typedef struct _McdAccountClass McdAccountClass;
 
-typedef struct _McdAccountPresencePrivate McdAccountPresencePrivate;
-
 #include "mcd-connection.h"
 #include "mcd-account-manager.h"
 
@@ -48,7 +46,6 @@ struct _McdAccount
 {
     GObject parent;
     McdAccountPrivate *priv;
-    McdAccountPresencePrivate *presence_priv;
 };
 
 typedef enum
@@ -67,37 +64,21 @@ typedef void (*McdAccountLoadCb) (McdAccount *account,
 typedef void (*McdAccountDeleteCb) (McdAccount *account,
                                     const GError *error,
                                     gpointer user_data);
-typedef void (*McdAccountSetParameterCb) (McdAccount *account,
-                                          const GError *error,
-                                          gpointer user_data);
-typedef void (*McdAccountGetParameterCb) (McdAccount *account,
-                                          const GValue *value,
-                                          const GError *error,
-                                          gpointer user_data);
 
 struct _McdAccountClass
 {
     GObjectClass parent_class;
-    void (*get_parameter) (McdAccount *account, const gchar *name,
-                           McdAccountGetParameterCb callback,
-                           gpointer user_data);
-    void (*set_parameter) (McdAccount *account, const gchar *name,
-                           const GValue *value,
-                           McdAccountSetParameterCb callback,
-                           gpointer user_data);
-    void (*delete) (McdAccount *account, McdAccountDeleteCb callback,
-                    gpointer user_data);
-    void (*load) (McdAccount *account, McdAccountLoadCb callback,
-                  gpointer user_data);
+    /* Used to be get_parameter; set_parameter; delete; load. These are not
+     * implementeed in any known subclass of this class, so have been removed;
+     * these padding fields are to preserve ABI compatibility.
+     */
+    GCallback dummy[4];
     gboolean (*check_request) (McdAccount *account, GHashTable *request,
                                GError **error);
     void (*_mc_reserved6) (void);
     void (*_mc_reserved7) (void);
 };
 
-
-#define MC_ACCOUNT_DBUS_OBJECT_BASE "/org/freedesktop/Telepathy/Account/"
-
 GType mcd_account_get_type (void);
 McdAccount *mcd_account_new (McdAccountManager *account_manager,
 			     const gchar *name);
@@ -123,6 +104,7 @@ gboolean mcd_account_is_enabled (McdAccount *account);
 
 const gchar *mcd_account_get_manager_name (McdAccount *account);
 const gchar *mcd_account_get_protocol_name (McdAccount *account);
+TpConnectionManager *mcd_account_get_cm (McdAccount *account);
 
 void mcd_account_request_presence (McdAccount *account,
 				   TpConnectionPresenceType type,
