@@ -3,10 +3,8 @@
 /*
  * This file is part of mission-control
  *
- * Copyright © 2007-2009 Nokia Corporation.
- * Copyright © 2009-2010 Collabora Ltd.
- *
- * Contact: Naba Kumar  <naba.kumar@nokia.com>
+ * Copyright © 2007-2011 Nokia Corporation.
+ * Copyright © 2009-2011 Collabora Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -36,6 +34,7 @@
  */
 
 #include "mcd-connection.h"
+#include "mcd-connection-service-points.h"
 
 #include <string.h>
 #include <sys/types.h>
@@ -67,6 +66,7 @@
 #include "mcd-channel.h"
 #include "mcd-provisioning-factory.h"
 #include "mcd-misc.h"
+#include "mcd-slacker.h"
 #include "sp_timestamp.h"
 
 #include "mcd-signals-marshal.h"
@@ -74,6 +74,8 @@
 #include "_gen/cli-Connection_Interface_Contact_Capabilities.h"
 #include "_gen/cli-Connection_Interface_Contact_Capabilities_Draft1-body.h"
 #include "_gen/cli-Connection_Interface_Contact_Capabilities-body.h"
+#include "_gen/cli-Connection_Interface_Power_Saving.h"
+#include "_gen/cli-Connection_Interface_Power_Saving-body.h"
 
 #define INITIAL_RECONNECTION_TIME   3 /* seconds */
 #define RECONNECTION_MULTIPLIER     3
@@ -124,6 +126,7 @@ struct _McdConnectionPrivate
     guint has_capabilities_if : 1;
     guint has_contact_capabilities_draft1_if : 1;
     guint has_contact_capabilities_if : 1;
+    guint has_power_saving_if : 1;
 
     /* FALSE until the dispatcher has said it's ready for us */
     guint dispatching_started : 1;
@@ -146,7 +149,15 @@ struct _McdConnectionPrivate
     gchar *alias;
 
     gboolean is_disposed;
-    
+    gboolean service_points_watched;
+
+    McdSlacker *slacker;
+
+    struct
+    {
+        TpIntSet *handles;
+        GSList *numbers;
+    } emergency;
 };
 
 typedef struct
@@ -164,6 +175,7 @@ enum
     PROP_TP_CONNECTION,
     PROP_ACCOUNT,
     PROP_DISPATCHER,
+    PROP_SLACKER,
 };
 
 enum
@@ -186,7 +198,6 @@ static const gchar * const *presence_fallbacks[] = {
     _available_fb, _away_fb, _ext_away_fb, _hidden_fb, _busy_fb
 };
 
-static GError * map_tp_error_to_mc_error (McdChannel *channel, const GError *tp_error);
 static void _mcd_connection_release_tp_connection (McdConnection *connection);
 static gboolean request_channel_new_iface (McdConnection *connection,
                                            McdChannel *channel);
@@ -411,8 +422,8 @@ presence_get_statuses_cb (TpProxy *proxy, const GValue *v_statuses,
     }
 
     /* Now the presence info is ready. We can set the presence */
-    _mcd_account_get_combined_presence (priv->account, &presence,
-                                        &status, &message);
+    _mcd_account_get_requested_presence (priv->account, &presence,
+                                         &status, &message);
     if (priv->connected)
     {
         priv->presence_info_ready = TRUE;
@@ -1058,6 +1069,21 @@ _mcd_connection_setup_alias (McdConnection *connection)
     g_array_free (self_handle_array, TRUE);
 }
 
+static void
+_mcd_connection_setup_power_saving (McdConnection *connection)
+{
+  McdConnectionPrivate *priv = connection->priv;
+
+  if (priv->slacker == NULL)
+    return;
+
+  DEBUG ("is %sactive", mcd_slacker_is_inactive (priv->slacker) ? "in" : "");
+
+  if (mcd_slacker_is_inactive (priv->slacker))
+    mc_cli_connection_interface_power_saving_call_set_power_saving (priv->tp_conn, -1,
+        TRUE, NULL, NULL, NULL, NULL);
+}
+
 static gboolean
 mcd_connection_reconnect (McdConnection *connection)
 {
@@ -1138,6 +1164,10 @@ on_connection_status_changed (TpConnection *tp_conn, GParamSpec *pspec,
                 priv->probation_drop_count = 0;
             }
 
+            mcd_connection_service_point_setup (connection,
+                                                !priv->service_points_watched);
+            priv->service_points_watched = TRUE;
+
             priv->connected = TRUE;
         }
         break;
@@ -1255,8 +1285,8 @@ request_unrequested_channels (McdConnection *connection)
     }
 }
 
-static McdChannel *
-find_channel_by_path (McdConnection *connection,
+McdChannel *
+mcd_connection_find_channel_by_path (McdConnection *connection,
                       const gchar *object_path)
 {
     const GList *list = NULL;
@@ -1339,7 +1369,7 @@ on_new_channels (TpConnection *proxy, const GPtrArray *channels,
 
         /* if the channel was a request, we already have an object for it;
          * otherwise, create a new one */
-        channel = find_channel_by_path (connection, object_path);
+        channel = mcd_connection_find_channel_by_path (connection, object_path);
         if (!channel)
         {
             channel = mcd_channel_new_from_properties (proxy, object_path,
@@ -1604,6 +1634,8 @@ on_connection_ready (TpConnection *tp_conn, const GError *error,
         MC_IFACE_QUARK_CONNECTION_INTERFACE_CONTACT_CAPABILITIES_DRAFT1);
     priv->has_contact_capabilities_if = tp_proxy_has_interface_by_id (tp_conn,
         TP_IFACE_QUARK_CONNECTION_INTERFACE_CONTACT_CAPABILITIES);
+    priv->has_power_saving_if = tp_proxy_has_interface_by_id (tp_conn,
+        MC_IFACE_QUARK_CONNECTION_INTERFACE_POWER_SAVING);
 
     if (priv->has_presence_if)
 	_mcd_connection_setup_presence (connection);
@@ -1620,6 +1652,9 @@ on_connection_ready (TpConnection *tp_conn, const GError *error,
     if (priv->has_alias_if)
 	_mcd_connection_setup_alias (connection);
 
+    if (priv->has_power_saving_if)
+      _mcd_connection_setup_power_saving (connection);
+
     if (!priv->dispatching_started)
         _mcd_dispatcher_add_connection (priv->dispatcher, connection);
 
@@ -2017,6 +2052,31 @@ on_account_removed (McdAccount *account, McdConnection *connection)
     mcd_mission_abort (MCD_MISSION (connection));
 }
 
+static void
+on_inactivity_changed (McdSlacker *slacker,
+    gboolean inactive,
+    McdConnection *self)
+{
+  McdConnectionPrivate *priv = self->priv;
+  DEBUG ("%sactive, %s have power saving iface.", inactive ? "in" : "",
+      priv->has_power_saving_if ? "has" : "doesn't");
+
+  if (priv->has_power_saving_if)
+    mc_cli_connection_interface_power_saving_call_set_power_saving (priv->tp_conn, -1,
+        inactive, NULL, NULL, NULL, NULL);
+}
+
+static void
+_mcd_connection_constructed (GObject * object)
+{
+    McdConnection *self = MCD_CONNECTION (object);
+    McdConnectionPrivate *priv = self->priv;
+
+    if (priv->slacker != NULL)
+      g_signal_connect (priv->slacker, "inactivity-changed",
+          G_CALLBACK (on_inactivity_changed), self);
+}
+
 static void
 _mcd_connection_dispose (GObject * object)
 {
@@ -2058,6 +2118,15 @@ _mcd_connection_dispose (GObject * object)
         tp_clear_object (&priv->account);
     }
 
+    if (priv->slacker != NULL)
+      {
+        g_signal_handlers_disconnect_by_func (priv->slacker,
+                                              G_CALLBACK (on_inactivity_changed),
+                                              connection);
+
+        tp_clear_object (&priv->slacker);
+      }
+
     tp_clear_object (&priv->tp_conn_mgr);
     tp_clear_object (&priv->dispatcher);
     tp_clear_object (&priv->dbus_daemon);
@@ -2106,6 +2175,10 @@ _mcd_connection_set_property (GObject * obj, guint prop_id,
                           obj);
         _mcd_account_set_connection (account, MCD_CONNECTION (obj));
 	break;
+    case PROP_SLACKER:
+      g_assert (priv->slacker == NULL);
+      priv->slacker = g_value_dup_object (val);
+    break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
 	break;
@@ -2135,6 +2208,9 @@ _mcd_connection_get_property (GObject * obj, guint prop_id,
     case PROP_DISPATCHER:
 	g_value_set_object (val, priv->dispatcher);
 	break;
+    case PROP_SLACKER:
+      g_value_set_object (val, priv->slacker);
+	break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
 	break;
@@ -2177,7 +2253,7 @@ mcd_connection_need_dispatch (McdConnection *connection,
         {
             any_requested = TRUE;
 
-            if (find_channel_by_path (connection, object_path))
+            if (mcd_connection_find_channel_by_path (connection, object_path))
                 requested_by_us = TRUE;
         }
     }
@@ -2187,9 +2263,60 @@ mcd_connection_need_dispatch (McdConnection *connection,
     return !any_requested || requested_by_us;
 }
 
+gboolean
+_mcd_connection_target_id_is_urgent (McdConnection *self, const gchar *name)
+{
+  GSList *list = self->priv->emergency.numbers;
+
+  for (; list != NULL; list = g_slist_next (list))
+    {
+      const gchar **number = list->data;
+
+      for (; number != NULL && *number != NULL; number++)
+        {
+          if (!tp_strdiff (*number, name))
+            return TRUE;
+        }
+    }
+
+  return FALSE;
+}
+
+gboolean
+_mcd_connection_target_handle_is_urgent (McdConnection *self, guint handle)
+{
+  if (handle != 0 && self->priv->emergency.handles != NULL)
+    return tp_intset_is_member (self->priv->emergency.handles, handle);
+
+  return FALSE;
+}
+
+gboolean
+_mcd_connection_channel_is_urgent (McdConnection *self, McdChannel *channel)
+{
+    const gchar *name = NULL;
+
+    if (mcd_channel_get_handle_type (channel) != TP_HANDLE_TYPE_CONTACT)
+        return FALSE;
+
+    name = mcd_channel_get_name (channel);
+
+    if (tp_str_empty (name))
+    {
+        TpHandle handle = mcd_channel_get_handle (channel);
+
+        return _mcd_connection_target_handle_is_urgent (self, handle);
+    }
+    else
+    {
+        return _mcd_connection_target_id_is_urgent (self, name);
+    }
+
+    return FALSE;
+}
+
 static gboolean
-_mcd_connection_request_channel (McdConnection *connection,
-                                 McdChannel *channel)
+_mcd_connection_request_channel (McdConnection *connection, McdChannel *channel)
 {
     McdConnectionPrivate *priv = MCD_CONNECTION_PRIV (connection);
     gboolean ret;
@@ -2237,6 +2364,8 @@ mcd_connection_add_signals (TpProxy *self,
                                                                          data);
     mc_cli_Connection_Interface_Contact_Capabilities_add_signals (self, quark,
                                                                   proxy, data);
+    mc_cli_Connection_Interface_Power_Saving_add_signals (self, quark,
+        proxy, data);
 }
 
 static void
@@ -2247,6 +2376,7 @@ mcd_connection_class_init (McdConnectionClass * klass)
 
     object_class->finalize = _mcd_connection_finalize;
     object_class->dispose = _mcd_connection_dispose;
+    object_class->constructed = _mcd_connection_constructed;
     object_class->set_property = _mcd_connection_set_property;
     object_class->get_property = _mcd_connection_get_property;
 
@@ -2293,6 +2423,13 @@ mcd_connection_class_init (McdConnectionClass * klass)
                               "Account",
                               MCD_TYPE_ACCOUNT,
                               G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+    g_object_class_install_property
+        (object_class, PROP_SLACKER,
+         g_param_spec_object ("slacker",
+                              "MCE slacker",
+                              "Slacker object notifies us of user inactivity",
+                              MCD_TYPE_SLACKER,
+                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
     signals[SELF_PRESENCE_CHANGED] = g_signal_new ("self-presence-changed",
         G_OBJECT_CLASS_TYPE (klass),
@@ -2344,29 +2481,6 @@ mcd_connection_get_account (McdConnection * id)
     return priv->account;
 }
 
-static GError *
-map_tp_error_to_mc_error (McdChannel *channel, const GError *error)
-{
-    McError mc_error_code;
-
-    DEBUG ("Telepathy Error = %s", error->message);
-
-    /* Some TP errors might be a bit too generic for the UIs.
-     * With some guesswork, we can add more precise error reporting here.
-     */
-    if (mcd_channel_get_channel_type_quark (channel) ==
-	TP_IFACE_QUARK_CHANNEL_TYPE_STREAMED_MEDIA &&
-	error->code == TP_ERROR_NOT_AVAILABLE)
-    {
-	mc_error_code = MC_CONTACT_DOES_NOT_SUPPORT_VOICE_ERROR;
-    }
-    else
-        return g_error_copy (error);
-
-    return g_error_new (MC_ERROR, mc_error_code, "Telepathy Error: %s",
-                        error->message);
-}
-
 static void
 common_request_channel_cb (TpConnection *proxy, gboolean yours,
                            const gchar *channel_path, GHashTable *properties,
@@ -2383,7 +2497,7 @@ common_request_channel_cb (TpConnection *proxy, gboolean yours,
          * https://bugs.freedesktop.org/show_bug.cgi?id=15769 will be fixed
          * soon :-) */
         DEBUG ("got error: %s", error->message);
-        mc_error = map_tp_error_to_mc_error (channel, error);
+        mc_error = g_error_copy (error);
         mcd_channel_take_error (channel, mc_error);
         mcd_mission_abort ((McdMission *)channel);
         return;
@@ -2397,7 +2511,8 @@ common_request_channel_cb (TpConnection *proxy, gboolean yours,
     if (_mcd_channel_get_request_use_existing (channel))
     {
         McdChannel *existing;
-        existing = find_channel_by_path (connection, channel_path);
+        existing = mcd_connection_find_channel_by_path (connection,
+            channel_path);
         if (existing)
         {
             _mcd_dispatcher_add_channel_request (priv->dispatcher, existing,
@@ -2727,3 +2842,68 @@ _mcd_connection_presence_info_is_ready (McdConnection *self)
 
     return self->priv->presence_info_ready;
 }
+
+static void clear_emergency_handles (McdConnectionPrivate *priv)
+{
+  guint n_handles = tp_intset_size (priv->emergency.handles);
+
+  /* trawl through the handles and unref them */
+  if (n_handles > 0)
+    {
+      TpIntSetFastIter iter;
+      TpHandle *handles = g_new0 (TpHandle, n_handles);
+      TpHandle handle;
+      guint i = 0;
+
+      tp_intset_fast_iter_init (&iter, priv->emergency.handles);
+
+      while (tp_intset_fast_iter_next (&iter, &handle))
+        handles[i++] = handle;
+
+      tp_connection_unref_handles (priv->tp_conn, TP_HANDLE_TYPE_CONTACT,
+          n_handles, handles);
+
+      g_free (handles);
+    }
+
+  tp_clear_pointer (&priv->emergency.handles, tp_intset_destroy);
+}
+
+static void clear_emergency_numbers (McdConnectionPrivate *priv)
+{
+  g_slist_foreach (priv->emergency.numbers, (GFunc) g_strfreev, NULL);
+  tp_clear_pointer (&priv->emergency.numbers, g_slist_free);
+}
+
+void _mcd_connection_clear_emergency_data (McdConnection *self)
+{
+    clear_emergency_handles (self->priv);
+    clear_emergency_numbers (self->priv);
+}
+
+void _mcd_connection_take_emergency_numbers (McdConnection *self, GSList *numbers)
+{
+    McdConnectionPrivate *priv = self->priv;
+
+    if (priv->emergency.numbers != NULL)
+      {
+        clear_emergency_numbers (priv);
+        g_critical ("Overwriting old emergency numbers");
+      }
+
+    priv->emergency.numbers = numbers;
+}
+
+void
+_mcd_connection_take_emergency_handles (McdConnection *self, TpIntSet *handles)
+{
+    McdConnectionPrivate *priv = self->priv;
+
+    if (priv->emergency.handles != NULL)
+      {
+        clear_emergency_handles (priv);
+        g_critical ("Overwriting old emergency handles");
+      }
+
+    priv->emergency.handles = handles;
+}
