@@ -3,8 +3,8 @@
 /*
  * This file is part of mission-control
  *
- * Copyright (C) 2008-2009 Nokia Corporation.
- * Copyright (C) 2009 Collabora Ltd.
+ * Copyright (C) 2008-2010 Nokia Corporation.
+ * Copyright (C) 2009-2010 Collabora Ltd.
  *
  * Contact: Alberto Mardegan  <alberto.mardegan@nokia.com>
  *
@@ -46,12 +46,13 @@
 #include "channel-utils.h"
 #include "mcd-channel-priv.h"
 #include "mcd-dbusprop.h"
+#include "mcd-master-priv.h"
 #include "mcd-misc.h"
+#include "plugin-dispatch-operation.h"
+#include "plugin-loader.h"
 
 #include <libmcclient/mc-errors.h>
-
-#define MCD_CLIENT_BASE_NAME "org.freedesktop.Telepathy.Client."
-#define MCD_CLIENT_BASE_NAME_LEN (sizeof (MCD_CLIENT_BASE_NAME) - 1)
+#include "libmcclient/mc-gtypes.h"
 
 #define MCD_DISPATCH_OPERATION_PRIV(operation) (MCD_DISPATCH_OPERATION (operation)->priv)
 
@@ -232,6 +233,11 @@ struct _McdDispatchOperationPrivate
      * A reference is held for each pending observer. */
     gsize observers_pending;
 
+    /* The number of observers that are pending which have
+     * DelayApprovers=TRUE. This is used to know if
+     * AddDispatchOperation can be called yet. */
+    gsize delay_approver_observers_pending;
+
     /* The number of approvers that have not yet returned from
      * AddDispatchOperation. Until they have done so, we can't allow the
      * dispatch operation to finish. This is a client lock.
@@ -253,6 +259,10 @@ struct _McdDispatchOperationPrivate
     /* If TRUE, we've tried all the BypassApproval handlers, which happens
      * before we run approvers. */
     gboolean tried_handlers_before_approval;
+
+    McdPluginDispatchOperation *plugin_api;
+    gsize plugins_pending;
+    gboolean did_post_observer_actions;
 };
 
 static void _mcd_dispatch_operation_check_finished (
@@ -277,7 +287,8 @@ mcd_dispatch_operation_may_signal_finished (McdDispatchOperation *self)
 }
 
 static void
-_mcd_dispatch_operation_inc_observers_pending (McdDispatchOperation *self)
+_mcd_dispatch_operation_inc_observers_pending (McdDispatchOperation *self,
+    McdClientProxy *client)
 {
     g_return_if_fail (self->priv->result == NULL);
 
@@ -287,10 +298,14 @@ _mcd_dispatch_operation_inc_observers_pending (McdDispatchOperation *self)
            self->priv->observers_pending,
            self->priv->observers_pending + 1);
     self->priv->observers_pending++;
+
+    if (_mcd_client_proxy_get_delay_approvers (client))
+      self->priv->delay_approver_observers_pending++;
 }
 
 static void
-_mcd_dispatch_operation_dec_observers_pending (McdDispatchOperation *self)
+_mcd_dispatch_operation_dec_observers_pending (McdDispatchOperation *self,
+    McdClientProxy *client)
 {
     DEBUG ("%" G_GSIZE_FORMAT " -> %" G_GSIZE_FORMAT,
            self->priv->observers_pending,
@@ -298,6 +313,9 @@ _mcd_dispatch_operation_dec_observers_pending (McdDispatchOperation *self)
     g_return_if_fail (self->priv->observers_pending > 0);
     self->priv->observers_pending--;
 
+    if (_mcd_client_proxy_get_delay_approvers (client))
+      self->priv->delay_approver_observers_pending--;
+
     _mcd_dispatch_operation_check_finished (self);
     _mcd_dispatch_operation_check_client_locks (self);
     g_object_unref (self);
@@ -360,19 +378,66 @@ static void _mcd_dispatch_operation_close_as_undispatchable (
     McdDispatchOperation *self, const GError *error);
 static gboolean mcd_dispatch_operation_idle_run_approvers (gpointer p);
 static void mcd_dispatch_operation_set_channel_handled_by (
-    McdDispatchOperation *self, McdChannel *channel, const gchar *unique_name);
+    McdDispatchOperation *self, McdChannel *channel, const gchar *unique_name,
+    const gchar *well_known_name);
+static gboolean _mcd_dispatch_operation_handlers_can_bypass_approval (
+    McdDispatchOperation *self);
+static gboolean _mcd_dispatch_operation_handlers_can_bypass_observers (
+    McdDispatchOperation *self);
 
 static void
 _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
 {
     Approval *approval;
+    guint approver_event_id = 0;
+
+    if (!self->priv->invoked_observers_if_needed)
+    {
+        DEBUG ("waiting for Observers to be called");
+        return;
+    }
+
+    if (self->priv->plugins_pending > 0)
+    {
+        DEBUG ("waiting for plugins to stop delaying");
+        return;
+    }
+
+    /* Check whether plugins' requests to close channels later should be
+     * honoured. We want to do this before running Approvers (if any). */
+    if (self->priv->observers_pending == 0 &&
+        !self->priv->did_post_observer_actions)
+    {
+        _mcd_plugin_dispatch_operation_observers_finished (
+            self->priv->plugin_api);
+        self->priv->did_post_observer_actions = TRUE;
+    }
+
+    /* If nobody is bypassing approval, then we want to run approvers as soon
+     * as possible, without waiting for observers, to improve responsiveness.
+     * (The regression test dispatcher/exploding-bundles.py asserts that we
+     * do this.)
+     *
+     * However, if a handler bypasses approval, we must wait til the observers
+     * return, then run that handler, then proceed with the other handlers. */
+    if (!self->priv->tried_handlers_before_approval &&
+        !_mcd_dispatch_operation_handlers_can_bypass_approval (self)
+        && self->priv->delay_approver_observers_pending == 0
+        && self->priv->channels != NULL &&
+        ! _mcd_plugin_dispatch_operation_will_terminate (
+            self->priv->plugin_api))
+    {
+        self->priv->tried_handlers_before_approval = TRUE;
+
+        approver_event_id = g_idle_add_full (G_PRIORITY_HIGH,
+                         mcd_dispatch_operation_idle_run_approvers,
+                         g_object_ref (self), g_object_unref);
+    }
 
     /* we may not continue until we've called all the Observers, and they've
      * all replied "I'm ready" */
-    if (!self->priv->invoked_observers_if_needed ||
-        self->priv->observers_pending > 0)
+    if (self->priv->observers_pending > 0)
     {
-        DEBUG ("waiting for Observers");
         return;
     }
 
@@ -428,7 +493,7 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
             McdChannel *channel = MCD_CHANNEL (list->data);
 
             mcd_dispatch_operation_set_channel_handled_by (self, channel,
-                caller);
+                caller, NULL);
         }
 
         DEBUG ("Replying to Claim call from %s", caller);
@@ -442,8 +507,25 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
                                         caller);
         g_free (caller);
 
+        if (approver_event_id > 0)
+        {
+            DEBUG ("Cancelling call to approvers as dispatch operation has been Claimed");
+            g_source_remove (approver_event_id);
+        }
+
         return;
     }
+    else if (approval != NULL && approval->type == APPROVAL_TYPE_HANDLE_WITH)
+    {
+        /* We set this to TRUE so that the handlers are called. */
+        self->priv->invoked_approvers_if_needed = TRUE;
+
+        if (approver_event_id > 0)
+        {
+            DEBUG ("Cancelling call to approvers as dispatch operation has been HandledWith'd");
+            g_source_remove (approver_event_id);
+        }
+    }
 
     if (self->priv->invoked_approvers_if_needed)
     {
@@ -501,7 +583,7 @@ enum
  * Returns: the D-Bus object path of the Connection associated with @self,
  *    or "/" if none.
  */
-static const gchar *
+const gchar *
 _mcd_dispatch_operation_get_connection_path (McdDispatchOperation *self)
 {
     const gchar *path;
@@ -530,6 +612,30 @@ get_connection (TpSvcDBusProperties *self, const gchar *name, GValue *value)
             MCD_DISPATCH_OPERATION (self)));
 }
 
+const gchar *
+_mcd_dispatch_operation_get_cm_name (McdDispatchOperation *self)
+{
+    const gchar *ret;
+
+    g_return_val_if_fail (MCD_IS_DISPATCH_OPERATION (self), NULL);
+    g_return_val_if_fail (self->priv->account != NULL, NULL);
+    ret = mcd_account_get_manager_name (self->priv->account);
+    g_return_val_if_fail (ret != NULL, NULL);
+    return ret;
+}
+
+const gchar *
+_mcd_dispatch_operation_get_protocol (McdDispatchOperation *self)
+{
+    const gchar *ret;
+
+    g_return_val_if_fail (MCD_IS_DISPATCH_OPERATION (self), NULL);
+    g_return_val_if_fail (self->priv->account != NULL, NULL);
+    ret = mcd_account_get_protocol_name (self->priv->account);
+    g_return_val_if_fail (ret != NULL, NULL);
+    return ret;
+}
+
 /*
  * _mcd_dispatch_operation_get_account_path:
  * @operation: the #McdDispatchOperation.
@@ -537,7 +643,7 @@ get_connection (TpSvcDBusProperties *self, const gchar *name, GValue *value)
  * Returns: the D-Bus object path of the Account associated with @operation,
  *    or "/" if none.
  */
-static const gchar *
+const gchar *
 _mcd_dispatch_operation_get_account_path (McdDispatchOperation *self)
 {
     const gchar *path;
@@ -610,21 +716,20 @@ properties_iface_init (TpSvcDBusPropertiesClass *iface, gpointer iface_data)
 static void
 mcd_dispatch_operation_set_channel_handled_by (McdDispatchOperation *self,
                                                McdChannel *channel,
-                                               const gchar *unique_name)
+                                               const gchar *unique_name,
+                                               const gchar *well_known_name)
 {
-    const gchar *path;
     TpChannel *tp_channel;
 
     g_assert (unique_name != NULL);
 
-    path = mcd_channel_get_object_path (channel);
     tp_channel = mcd_channel_get_tp_channel (channel);
     g_return_if_fail (tp_channel != NULL);
 
     _mcd_channel_set_status (channel, MCD_CHANNEL_STATUS_DISPATCHED);
 
     _mcd_handler_map_set_channel_handled (self->priv->handler_map,
-        tp_channel, unique_name,
+        tp_channel, unique_name, well_known_name,
         _mcd_dispatch_operation_get_account_path (self));
 }
 
@@ -663,7 +768,7 @@ _mcd_dispatch_operation_finish (McdDispatchOperation *operation,
     }
 
     va_start (ap, format);
-    priv->result = _mcd_g_error_new_valist (domain, code, format, ap);
+    priv->result = g_error_new_valist (domain, code, format, ap);
     va_end (ap);
     DEBUG ("Result: %s", priv->result->message);
 
@@ -696,6 +801,9 @@ _mcd_dispatch_operation_finish (McdDispatchOperation *operation,
                     {
                         DEBUG ("successful HandleWith, channel went to %s",
                                successful_handler);
+
+                        /* HandleWith and HandleWithTime both return void, so
+                         * it's OK to not distinguish */
                         tp_svc_channel_dispatch_operation_return_from_handle_with (
                             approval->context);
                     }
@@ -745,9 +853,10 @@ static gboolean mcd_dispatch_operation_check_handle_with (
     McdDispatchOperation *self, const gchar *handler_name, GError **error);
 
 static void
-dispatch_operation_handle_with (TpSvcChannelDispatchOperation *cdo,
-                                const gchar *handler_name,
-                                DBusGMethodInvocation *context)
+dispatch_operation_handle_with_time (TpSvcChannelDispatchOperation *cdo,
+    const gchar *handler_name,
+    gint64 user_action_timestamp,
+    DBusGMethodInvocation *context)
 {
     GError *error = NULL;
     McdDispatchOperation *self = MCD_DISPATCH_OPERATION (cdo);
@@ -761,14 +870,23 @@ dispatch_operation_handle_with (TpSvcChannelDispatchOperation *cdo,
         return;
     }
 
-    /* 0 is a special case for 'no user action' */
-    self->priv->handle_with_time = 0;
+    self->priv->handle_with_time = user_action_timestamp;
 
     g_queue_push_tail (self->priv->approvals,
                        approval_new_handle_with (handler_name, context));
     _mcd_dispatch_operation_check_client_locks (self);
 }
 
+static void
+dispatch_operation_handle_with (TpSvcChannelDispatchOperation *cdo,
+    const gchar *handler_name,
+    DBusGMethodInvocation *context)
+{
+    dispatch_operation_handle_with_time (cdo, handler_name,
+                                         TP_USER_ACTION_TIME_NOT_USER_ACTION,
+                                         context);
+}
+
 static void
 dispatch_operation_claim (TpSvcChannelDispatchOperation *cdo,
                           DBusGMethodInvocation *context)
@@ -800,6 +918,7 @@ dispatch_operation_iface_init (TpSvcChannelDispatchOperationClass *iface,
     iface, dispatch_operation_##x)
     IMPLEMENT(handle_with);
     IMPLEMENT(claim);
+    IMPLEMENT(handle_with_time);
 #undef IMPLEMENT
 }
 
@@ -883,6 +1002,8 @@ mcd_dispatch_operation_constructor (GType type, guint n_params,
         g_object_unref (dbus_daemon);
     }
 
+    priv->plugin_api = _mcd_plugin_dispatch_operation_new (operation);
+
     return object;
 error:
     g_object_unref (object);
@@ -970,7 +1091,7 @@ mcd_dispatch_operation_set_property (GObject *obj, guint prop_id,
                 _mcd_channel_get_request_preferred_handler (channel);
 
             if (preferred_handler != NULL &&
-                g_str_has_prefix (preferred_handler, MCD_CLIENT_BASE_NAME) &&
+                g_str_has_prefix (preferred_handler, TP_CLIENT_BUS_NAME_BASE) &&
                 tp_dbus_check_valid_bus_name (preferred_handler,
                                               TP_DBUS_NAME_TYPE_WELL_KNOWN,
                                               NULL))
@@ -1065,23 +1186,10 @@ mcd_dispatch_operation_finalize (GObject *object)
 {
     McdDispatchOperationPrivate *priv = MCD_DISPATCH_OPERATION_PRIV (object);
 
-    g_strfreev (priv->possible_handlers);
-    priv->possible_handlers = NULL;
-
-    if (priv->properties)
-        g_hash_table_unref (priv->properties);
-
-    if (priv->failed_handlers != NULL)
-    {
-        g_hash_table_unref (priv->failed_handlers);
-    }
-
-    if (priv->result != NULL)
-    {
-        g_error_free (priv->result);
-        priv->result = NULL;
-    }
-
+    tp_clear_pointer (&priv->possible_handlers, g_strfreev);
+    tp_clear_pointer (&priv->properties, g_hash_table_unref);
+    tp_clear_pointer (&priv->failed_handlers, g_hash_table_unref);
+    g_clear_error (&priv->result);
     g_free (priv->object_path);
 
     G_OBJECT_CLASS (_mcd_dispatch_operation_parent_class)->finalize (object);
@@ -1093,13 +1201,10 @@ mcd_dispatch_operation_dispose (GObject *object)
     McdDispatchOperationPrivate *priv = MCD_DISPATCH_OPERATION_PRIV (object);
     GList *list;
 
-    if (priv->successful_handler != NULL)
-    {
-        g_object_unref (priv->successful_handler);
-        priv->successful_handler = NULL;
-    }
+    tp_clear_object (&priv->plugin_api);
+    tp_clear_object (&priv->successful_handler);
 
-    if (priv->channels)
+    if (priv->channels != NULL)
     {
         for (list = priv->channels; list != NULL; list = list->next)
         {
@@ -1108,47 +1213,26 @@ mcd_dispatch_operation_dispose (GObject *object)
             g_object_unref (list->data);
         }
 
-        g_list_free (priv->channels);
-        priv->channels = NULL;
+        tp_clear_pointer (&priv->channels, g_list_free);
     }
 
     if (priv->lost_channels != NULL)
     {
         for (list = priv->lost_channels; list != NULL; list = list->next)
             g_object_unref (list->data);
-        g_list_free (priv->lost_channels);
-        priv->lost_channels = NULL;
+
+        tp_clear_pointer (&priv->lost_channels, g_list_free);
     }
 
-    if (priv->connection)
-    {
-        g_object_unref (priv->connection);
-        priv->connection = NULL;
-    }
-
-    if (priv->account != NULL)
-    {
-        g_object_unref (priv->account);
-        priv->account = NULL;
-    }
-
-    if (priv->handler_map != NULL)
-    {
-        g_object_unref (priv->handler_map);
-        priv->handler_map = NULL;
-    }
-
-    if (priv->client_registry != NULL)
-    {
-        g_object_unref (priv->client_registry);
-        priv->client_registry = NULL;
-    }
+    tp_clear_object (&priv->connection);
+    tp_clear_object (&priv->account);
+    tp_clear_object (&priv->handler_map);
+    tp_clear_object (&priv->client_registry);
 
     if (priv->approvals != NULL)
     {
         g_queue_foreach (priv->approvals, (GFunc) approval_free, NULL);
-        g_queue_free (priv->approvals);
-        priv->approvals = NULL;
+        tp_clear_pointer (&priv->approvals, g_queue_free);
     }
 
     G_OBJECT_CLASS (_mcd_dispatch_operation_parent_class)->dispose (object);
@@ -1376,7 +1460,7 @@ mcd_dispatch_operation_check_handle_with (McdDispatchOperation *self,
         return TRUE;
     }
 
-    if (!g_str_has_prefix (handler_name, MCD_CLIENT_BASE_NAME) ||
+    if (!g_str_has_prefix (handler_name, TP_CLIENT_BUS_NAME_BASE) ||
         !tp_dbus_check_valid_bus_name (handler_name,
                                        TP_DBUS_NAME_TYPE_WELL_KNOWN, NULL))
     {
@@ -1402,7 +1486,7 @@ _mcd_dispatch_operation_approve (McdDispatchOperation *self,
     DEBUG ("%s/%p (preferred handler: '%s')", self->priv->unique_name, self,
            preferred_handler);
 
-    if (!g_str_has_prefix (preferred_handler, MCD_CLIENT_BASE_NAME) ||
+    if (!g_str_has_prefix (preferred_handler, TP_CLIENT_BUS_NAME_BASE) ||
         !tp_dbus_check_valid_bus_name (preferred_handler,
                                        TP_DBUS_NAME_TYPE_WELL_KNOWN, NULL))
     {
@@ -1642,6 +1726,34 @@ _mcd_dispatch_operation_handlers_can_bypass_approval (
     return FALSE;
 }
 
+/* this is analogous to *_can_bypass_handlers() method above */
+static gboolean
+_mcd_dispatch_operation_handlers_can_bypass_observers (
+    McdDispatchOperation *self)
+{
+    gchar **iter;
+
+    for (iter = self->priv->possible_handlers;
+         iter != NULL && *iter != NULL;
+         iter++)
+    {
+        McdClientProxy *handler = _mcd_client_registry_lookup (
+            self->priv->client_registry, *iter);
+
+        if (handler != NULL)
+        {
+            gboolean bypass = _mcd_client_proxy_get_bypass_observers (
+                handler);
+
+            DEBUG ("%s has BypassObservers=%c", *iter, bypass ? 'T' : 'F');
+            return bypass;
+        }
+    }
+
+    return FALSE;
+}
+
+
 gboolean
 _mcd_dispatch_operation_has_channel (McdDispatchOperation *self,
                                      McdChannel *channel)
@@ -1722,7 +1834,7 @@ _mcd_dispatch_operation_handle_channels_cb (TpClient *client,
             }
 
             mcd_dispatch_operation_set_channel_handled_by (self, channel,
-                                                           unique_name);
+                unique_name, tp_proxy_get_bus_name (client));
         }
 
         /* emit Finished, if we haven't already; but first make a note of the
@@ -1751,46 +1863,64 @@ observe_channels_cb (TpClient *proxy, const GError *error,
     else
         DEBUG ("success from %s", tp_proxy_get_object_path (proxy));
 
-    _mcd_dispatch_operation_dec_observers_pending (self);
+    _mcd_dispatch_operation_dec_observers_pending (self, MCD_CLIENT_PROXY (proxy));
 }
 
-/* The returned GPtrArray is allocated, but the contents are borrowed. */
-static GPtrArray *
-collect_satisfied_requests (GList *channels)
+/*
+ * @paths_out: (out) (transfer container) (element-type utf8):
+ *  Requests_Satisfied
+ * @props_out: (out) (transfer container) (element-type utf8 GHashTable):
+ *  request-properties for Observer_Info or Handler_Info
+ */
+static void
+collect_satisfied_requests (const GList *channels,
+    GPtrArray **paths_out,
+    GHashTable **props_out)
 {
-    const GList *c, *r;
-    GHashTable *set = g_hash_table_new (g_str_hash, g_str_equal);
-    GHashTableIter iter;
-    gpointer path;
-    GPtrArray *ret;
+    const GList *c;
+    GHashTable *set;
+    GHashTableIter it;
+    gpointer path, value;
+    GPtrArray *satisfied_requests;
+    GHashTable *request_properties;
+
+    set = g_hash_table_new_full (g_str_hash, g_str_equal,
+        g_free, g_object_unref);
 
-    /* collect object paths into a hash table, to drop duplicates
-     * FIXME (fd.o #24763): this shouldn't be necessary, because there should
-     * never be duplicates, unless my analysis is wrong? */
     for (c = channels; c != NULL; c = c->next)
     {
-        const GList *reqs = _mcd_channel_get_satisfied_requests (c->data,
+        GHashTable *reqs = _mcd_channel_get_satisfied_requests (c->data,
                                                                  NULL);
-
-        for (r = reqs; r != NULL; r = r->next)
-        {
-            g_hash_table_insert (set, r->data, r->data);
-        }
+        tp_g_hash_table_update (set, reqs,
+            (GBoxedCopyFunc) g_strdup, (GBoxedCopyFunc) g_object_ref);
+        g_hash_table_unref (reqs);
     }
 
-    /* serialize them into a pointer array, which is what dbus-glib wants */
-    ret = g_ptr_array_sized_new (g_hash_table_size (set));
+    satisfied_requests = g_ptr_array_sized_new (g_hash_table_size (set));
+    request_properties = g_hash_table_new_full (g_str_hash, g_str_equal,
+        NULL, (GDestroyNotify) g_hash_table_unref);
 
-    g_hash_table_iter_init (&iter, set);
+    g_hash_table_iter_init (&it, set);
 
-    while (g_hash_table_iter_next (&iter, &path, NULL))
+    while (g_hash_table_iter_next (&it, &path, &value))
     {
-        g_ptr_array_add (ret, path);
+        GHashTable *props;
+
+        g_ptr_array_add (satisfied_requests, path);
+        props = _mcd_request_dup_immutable_properties (value);
+        g_assert (props != NULL);
+        g_hash_table_insert (request_properties, path, props);
     }
 
-    g_hash_table_destroy (set);
+    if (paths_out != NULL)
+        *paths_out = satisfied_requests;
+    else
+        g_ptr_array_unref (satisfied_requests);
 
-    return ret;
+    if (props_out != NULL)
+        *props_out = request_properties;
+    else
+        g_hash_table_unref (request_properties);
 }
 
 static void
@@ -1802,7 +1932,7 @@ _mcd_dispatch_operation_run_observers (McdDispatchOperation *self)
     GHashTableIter iter;
     gpointer client_p;
 
-    observer_info = g_hash_table_new (g_str_hash, g_str_equal);
+    observer_info = tp_asv_new (NULL, NULL);
 
     _mcd_client_registry_init_hash_iter (self->priv->client_registry, &iter);
 
@@ -1812,6 +1942,7 @@ _mcd_dispatch_operation_run_observers (McdDispatchOperation *self)
         GList *observed = NULL;
         const gchar *account_path, *connection_path;
         GPtrArray *channels_array, *satisfied_requests;
+        GHashTable *request_properties;
 
         if (!tp_proxy_has_interface_by_id (client,
                                            TP_IFACE_QUARK_CLIENT_OBSERVER))
@@ -1841,14 +1972,22 @@ _mcd_dispatch_operation_run_observers (McdDispatchOperation *self)
          * if the observed list is the same */
         channels_array = _mcd_tp_channel_details_build_from_list (observed);
 
-        satisfied_requests = collect_satisfied_requests (observed);
+        collect_satisfied_requests (observed, &satisfied_requests,
+                                    &request_properties);
+
+        /* transfer ownership into observer_info */
+        /* FIXME: use telepathy-glib type when available */
+        tp_asv_take_boxed (observer_info, "request-properties",
+            MC_HASH_TYPE_OBJECT_IMMUTABLE_PROPERTIES_MAP,
+            request_properties);
+        request_properties = NULL;
 
         if (_mcd_dispatch_operation_needs_approval (self))
         {
             dispatch_operation_path = _mcd_dispatch_operation_get_path (self);
         }
 
-        _mcd_dispatch_operation_inc_observers_pending (self);
+        _mcd_dispatch_operation_inc_observers_pending (self, client);
 
         DEBUG ("calling ObserveChannels on %s for CDO %p",
                tp_proxy_get_bus_name (client), self);
@@ -1861,7 +2000,7 @@ _mcd_dispatch_operation_run_observers (McdDispatchOperation *self)
 
         /* don't free the individual object paths, which are borrowed from the
          * McdChannel objects */
-        g_ptr_array_free (satisfied_requests, TRUE);
+        g_ptr_array_unref (satisfied_requests);
 
         _mcd_tp_channel_details_free (channels_array);
 
@@ -1999,29 +2138,32 @@ _mcd_dispatch_operation_run_clients (McdDispatchOperation *self)
 
     if (self->priv->channels != NULL)
     {
-        _mcd_dispatch_operation_run_observers (self);
+        const GList *mini_plugins;
+
+        if (_mcd_dispatch_operation_handlers_can_bypass_observers (self))
+        {
+            DEBUG ("Bypassing observers");
+        }
+        else
+        {
+            DEBUG ("Running observers");
+            _mcd_dispatch_operation_run_observers (self);
+        }
+
+        for (mini_plugins = mcp_list_objects ();
+             mini_plugins != NULL;
+             mini_plugins = mini_plugins->next)
+        {
+            if (MCP_IS_DISPATCH_OPERATION_POLICY (mini_plugins->data))
+            {
+                mcp_dispatch_operation_policy_check (mini_plugins->data,
+                    MCP_DISPATCH_OPERATION (self->priv->plugin_api));
+            }
+        }
     }
 
     DEBUG ("All necessary observers invoked");
     self->priv->invoked_observers_if_needed = TRUE;
-    /* we call check_finished before returning */
-
-    /* If nobody is bypassing approval, then we want to run approvers as soon
-     * as possible, without waiting for observers, to improve responsiveness.
-     * (The regression test dispatcher/exploding-bundles.py asserts that we
-     * do this.)
-     *
-     * However, if a handler bypasses approval, we must wait til the observers
-     * return, then run that handler, then proceed with the other handlers. */
-    if (!_mcd_dispatch_operation_handlers_can_bypass_approval (self)
-        && self->priv->channels != NULL)
-    {
-        self->priv->tried_handlers_before_approval = TRUE;
-
-        g_idle_add_full (G_PRIORITY_HIGH,
-                         mcd_dispatch_operation_idle_run_approvers,
-                         g_object_ref (self), g_object_unref);
-    }
 
     DEBUG ("Checking finished/locks");
     _mcd_dispatch_operation_check_finished (self);
@@ -2041,13 +2183,25 @@ static void
 mcd_dispatch_operation_handle_channels (McdDispatchOperation *self,
                                         McdClientProxy *handler)
 {
+    GHashTable *handler_info;
+    GHashTable *request_properties;
+
     g_assert (!self->priv->calling_handle_channels);
     self->priv->calling_handle_channels = TRUE;
 
+    handler_info = tp_asv_new (NULL, NULL);
+    collect_satisfied_requests (self->priv->channels, NULL,
+                                &request_properties);
+    tp_asv_take_boxed (handler_info, "request-properties",
+        MC_HASH_TYPE_OBJECT_IMMUTABLE_PROPERTIES_MAP, request_properties);
+    request_properties = NULL;
+
     _mcd_client_proxy_handle_channels (handler,
         -1, self->priv->channels, self->priv->handle_with_time,
-        NULL, _mcd_dispatch_operation_handle_channels_cb,
+        handler_info, _mcd_dispatch_operation_handle_channels_cb,
         g_object_ref (self), g_object_unref, NULL);
+
+    g_hash_table_unref (handler_info);
 }
 
 static gboolean
@@ -2154,3 +2308,99 @@ _mcd_dispatch_operation_close_as_undispatchable (McdDispatchOperation *self,
 
     g_list_free (channels);
 }
+
+void
+_mcd_dispatch_operation_start_plugin_delay (McdDispatchOperation *self)
+{
+    g_object_ref (self);
+    DEBUG ("%" G_GSIZE_FORMAT " -> %" G_GSIZE_FORMAT,
+           self->priv->plugins_pending,
+           self->priv->plugins_pending + 1);
+    self->priv->plugins_pending++;
+}
+
+void
+_mcd_dispatch_operation_end_plugin_delay (McdDispatchOperation *self)
+{
+    DEBUG ("%" G_GSIZE_FORMAT " -> %" G_GSIZE_FORMAT,
+           self->priv->plugins_pending,
+           self->priv->plugins_pending - 1);
+    g_return_if_fail (self->priv->plugins_pending > 0);
+    self->priv->plugins_pending--;
+
+    _mcd_dispatch_operation_check_client_locks (self);
+    g_object_unref (self);
+}
+
+void
+_mcd_dispatch_operation_forget_channels (McdDispatchOperation *self)
+{
+    /* make a temporary copy, which is destroyed during the loop - otherwise
+     * we'll be trying to iterate over the list at the same time
+     * that mcd_mission_abort results in modifying it, which would be bad */
+    GList *list = _mcd_dispatch_operation_dup_channels (self);
+
+    while (list != NULL)
+    {
+        mcd_mission_abort (list->data);
+        g_object_unref (list->data);
+        list = g_list_delete_link (list, list);
+    }
+
+    /* There should now be none left (they all aborted) */
+    g_return_if_fail (self->priv->channels == NULL);
+}
+
+void
+_mcd_dispatch_operation_leave_channels (McdDispatchOperation *self,
+                                        TpChannelGroupChangeReason reason,
+                                        const gchar *message)
+{
+    GList *list;
+
+    if (message == NULL)
+    {
+        message = "";
+    }
+
+    list = _mcd_dispatch_operation_dup_channels (self);
+
+    while (list != NULL)
+    {
+        _mcd_channel_depart (list->data, reason, message);
+        g_object_unref (list->data);
+        list = g_list_delete_link (list, list);
+    }
+
+    _mcd_dispatch_operation_forget_channels (self);
+}
+
+void
+_mcd_dispatch_operation_close_channels (McdDispatchOperation *self)
+{
+    GList *list = _mcd_dispatch_operation_dup_channels (self);
+
+    while (list != NULL)
+    {
+        _mcd_channel_close (list->data);
+        g_object_unref (list->data);
+        list = g_list_delete_link (list, list);
+    }
+
+    _mcd_dispatch_operation_forget_channels (self);
+}
+
+void
+_mcd_dispatch_operation_destroy_channels (McdDispatchOperation *self)
+{
+    GList *list = _mcd_dispatch_operation_dup_channels (self);
+
+    while (list != NULL)
+    {
+        _mcd_channel_undispatchable (list->data);
+        g_object_unref (list->data);
+        list = g_list_delete_link (list, list);
+    }
+
+    _mcd_dispatch_operation_forget_channels (self);
+}
