@@ -31,9 +31,6 @@
 
 #include <telepathy-glib/telepathy-glib.h>
 
-#include <libmcclient/_gen/interfaces.h>
-#include <libmcclient/mc-profile.h>
-
 static gchar *app_name;
 static GMainLoop *main_loop;
 
@@ -45,9 +42,8 @@ show_help (gchar * err)
 
     printf ("Usage:\n"
 	    "    %1$s list\n"
-	    "    %1$s add <profile> <display name> [(int|bool|string):<key>=<value> ...]\n"
-	    "    %1$s add <manager>/<protocol> <display name> [(int|bool|string):<key>=<value> ...]\n"
-	    "    %1$s update <account name> [(int|uint|bool|string):<key>=<value>|clear:key] ...\n"
+	    "    %1$s add <manager>/<protocol> <display name> [<param> ...]\n"
+	    "    %1$s update <account name> [<param>|clear:key] ...\n"
 	    "    %1$s display <account name> <display name>\n"
 	    "    %1$s nick <account name> <nick name>\n"
 	    "    %1$s service <account name> <service name>\n"
@@ -59,7 +55,8 @@ show_help (gchar * err)
 	    "    %1$s auto-presence <account name> <presence status> [<message>]\n"
 	    "    %1$s request <account name> <presence status> [<message>]\n"
 	    "    %1$s auto-connect <account name> [(on|off)]\n"
-	    "    %1$s remove <account name>\n",
+	    "    %1$s remove <account name>\n"
+	    "  where <param> matches (int|uint|bool|string|path):<key>=<value>\n",
 	    app_name);
 
     if (err)
@@ -84,7 +81,7 @@ union command {
 
     struct {
 	struct common common;
-	gchar const *manager, *protocol, *profile, *display;
+	gchar const *manager, *protocol, *display;
 	GHashTable *parameters;
     } add;
 
@@ -233,6 +230,12 @@ set_param (GHashTable *parameters,
 	g_value_set_string (gvalue, value);
 	ret = TRUE;
     }
+    else if (strcmp (type, "path") == 0)
+    {
+	g_value_init (gvalue, DBUS_TYPE_G_OBJECT_PATH);
+	g_value_set_boxed (gvalue, value);
+	ret = TRUE;
+    }
 
     if (ret)
 	g_hash_table_replace (parameters, g_strdup (key), gvalue);
@@ -268,6 +271,10 @@ show_param (gchar const *key, GValue *value)
 	type = "bool";
 	decoded = g_strdup (g_value_get_boolean (value) ? "true" : "false");
     }
+    else if (G_VALUE_HOLDS (value, DBUS_TYPE_G_OBJECT_PATH)) {
+	type = "path";
+	decoded = g_value_dup_boxed (value);
+    }
     else {
 	type = G_VALUE_TYPE_NAME (value);
 	decoded = g_strdup_value_contents (value);
@@ -535,19 +542,9 @@ static gboolean
 command_add (TpAccountManager *manager)
 {
     GHashTable *properties;
-    GValue v_profile = { 0 };
 
     properties = g_hash_table_new (g_str_hash, g_str_equal);
 
-    if (command.add.profile)
-    {
-	g_value_init (&v_profile, G_TYPE_STRING);
-	g_value_set_static_string (&v_profile, command.add.profile);
-	g_hash_table_insert (properties,
-			     MC_IFACE_ACCOUNT_INTERFACE_COMPAT ".Profile",
-			     &v_profile);
-    }
-
     return NULL !=
 	tp_cli_account_manager_call_create_account
 	    (manager, 25000,
@@ -644,6 +641,7 @@ command_show (TpAccount *account)
     show ("Connects",
 	  tp_account_get_connect_automatically (account) ? "automatically" : NULL);
     show ("Nickname", tp_account_get_nickname (account));
+    show ("Service", tp_account_get_service (account));
 
     automatic.type = account_get_automatic_presence (account,
                                                      &automatic.status,
@@ -929,19 +927,9 @@ parse (int argc, char **argv)
 	}
 	else
 	{
-	    McProfile *profile;
-
-	    profile = mc_profile_lookup (argv[2]);
-	    if (!profile)
-	    {
-		g_warning ("%s: profile %s not found", argv[1], argv[2]);
-		exit (1);
-	    }
-
-	    command.add.profile = argv[2];
-	    command.add.manager = mc_profile_get_manager_name (profile);
-	    command.add.protocol = mc_profile_get_protocol_name (profile);
+	    show_help ("Invalid add command.");
 	}
+
 	command.ready.manager = command_add;
 	command.add.display = argv[3];
 
@@ -1256,10 +1244,10 @@ void account_ready (GObject *account,
     GError *error = NULL;
 
     if (!tp_proxy_prepare_finish (account, res, &error)) {
-	fprintf (stderr, "%s: %s: %s\n",
-		 app_name, skip_prefix (command.common.account),
-                 error->message);
-	g_error_free (error);
+        fprintf (stderr, "%s: couldn't load account '%s': %s\n", app_name,
+            skip_prefix (command.common.account), error->message);
+        fprintf (stderr, "Try '%s list' to list known accounts.\n", app_name);
+        g_error_free (error);
     }
     else {
         /* not all properties are exposed through TpAccount, request the
